import pytest
import time

from unoserver import server


@pytest.fixture(scope="session")
def server_fixture():
    srvr = server.UnoServer()
    process = srvr.start()
    # Give libreoffice a chance to start
    time.sleep(8)
    yield process  # provide the fixture value
    print("Teardown Unoserver")
    process.terminate()



"""Unoconverter unit tests"""
import os
import pytest
import sys
import uno

from unoserver import converter

TEST_DOCS = os.path.join(os.path.abspath(os.path.split(__file__)[0]), "documents")


def test_get_doc_type():
    ctx = uno.getComponentContext()
    smgr = ctx.ServiceManager
    # Pass in the service manager, it doesn't support any known document type
    with pytest.raises(RuntimeError):
        converter.get_doc_type(smgr)


old_import = __builtins__["__import__"]


def new_import(name, *optargs, **kwargs):
    if name == "uno":
        raise ImportError
    else:
        return old_import(name, *optargs, **kwargs)


def test_no_uno(monkeypatch):
    # Patch the import. Mock doesn't work here, because it refuses to
    # deal with __builtins__ as a module, so We can't use mock.patch().
    # Pytests monkeypatch will unmonkeypatch at the end of the test.
    monkeypatch.setitem(__builtins__, "__import__", new_import)
    del sys.modules["uno"]
    del sys.modules["unoserver"]
    del sys.modules["unoserver.converter"]

    # This should now raise an import error
    with pytest.raises(ImportError) as e:
        from unoserver import converter  # noqa: F401

    assert "This package must be installed with a Python" in str(e.value)


def test_wrong_arguments(monkeypatch):

    monkeypatch.setattr(converter.UnoConverter, "__init__", lambda self: None)
    conv = converter.UnoConverter()

    with pytest.raises(RuntimeError):
        # You need to pass in an infile, or data
        conv.convert()

    with pytest.raises(RuntimeError):
        # But not both
        conv.convert(inpath="somesortoffile.xls", indata="Shoobx rules!")

    with pytest.raises(RuntimeError):
        # You need to pass in an outfile or a file type
        conv.convert(inpath="somesortoffile.xls")



"""Tests that start a real unoserver and does real things"""

import io
import os
import pytest
import sys
import tempfile
import time

from unoserver import converter, server


TEST_DOCS = os.path.join(os.path.abspath(os.path.split(__file__)[0]), "documents")


@pytest.mark.parametrize("filename", ["simple.odt", "simple.xlsx"])
def test_pdf_conversion(server_fixture, filename):
    infile = os.path.join(TEST_DOCS, filename)

    with tempfile.NamedTemporaryFile(suffix=".pdf") as outfile:
        # Let Libreoffice write to the file and close it.
        sys.argv = ["unoconverter", infile, outfile.name]
        converter.main()

        # We now open it to check it, we can't use the outfile object,
        # it won't reflect the external changes.
        with open(outfile.name, "rb") as testfile:
            start = testfile.readline()
            assert start == b"%PDF-1.5\n"


class FakeStdio(io.BytesIO):
    """A BytesIO with a buffer attribute, usable to send binary stdin data"""

    @property
    def buffer(self):
        return self


@pytest.mark.parametrize("filename", ["simple.odt", "simple.xlsx"])
def test_stdin_stdout(server_fixture, monkeypatch, filename):

    with open(os.path.join(TEST_DOCS, filename), "rb") as infile:
        infile_stream = FakeStdio(infile.read())

    outfile_stream = FakeStdio()

    monkeypatch.setattr("sys.stdin", infile_stream)
    monkeypatch.setattr("sys.stdout", outfile_stream)

    sys.argv = ["unoconverter", "-", "-", "--convert-to", "pdf"]
    converter.main()

    outfile_stream.seek(0)
    start = outfile_stream.readline()
    assert start == b"%PDF-1.5\n"


def test_csv_conversion(server_fixture):
    conv = converter.UnoConverter()
    infile = os.path.join(TEST_DOCS, "simple.xlsx")

    with tempfile.NamedTemporaryFile(suffix=".csv") as outfile:
        # Let Libreoffice write to the file and close it.
        conv.convert(inpath=infile, outpath=outfile.name)
        # We now open it to check it, we can't use the outfile object,
        # it won't reflect the external changes.
        with open(outfile.name, "rb") as testfile:
            contents = testfile.readline()
            assert contents == b"This,Is,A,Simple,Excel,File\n"
            contents = testfile.readline()
            assert contents == b"1,2,3,4,5,6\n"


def test_impossible_conversion(server_fixture):
    conv = converter.UnoConverter()
    infile = os.path.join(TEST_DOCS, "simple.odt")

    with tempfile.NamedTemporaryFile(suffix=".xls") as outfile:
        # Let Libreoffice write to the file and close it.
        with pytest.raises(RuntimeError) as e:
            conv.convert(inpath=infile, outpath=outfile.name)
            assert "Could not find an export filter" in e


def test_multiple_servers(server_fixture):
    # The server fixture should already have started a server.
    # Make sure we can start a second one.
    sys.argv = ["unoserver", "--daemon"]
    process = server.main()
    try:
        # Wait for it to start
        time.sleep(5)
        # Make sure the process is still running, meaning return_code is None
        assert process.returncode is None
    finally:
        # Now kill the process
        process.terminate()
        # Wait for it to terminate
        process.wait()
        # And verify that it was killed
        assert process.returncode == 255


def test_unknown_outfile_type(server_fixture):
    infile = os.path.join(TEST_DOCS, "simple.odt")

    with tempfile.NamedTemporaryFile(suffix=".bog") as outfile:
        sys.argv = ["unoconverter", infile, outfile.name]
        # Type detection should fail, as it's not a .doc file:
        with pytest.raises(RuntimeError):
            converter.main()



import argparse
import logging
import os
import signal
import subprocess
import tempfile
from urllib import request

logger = logging.getLogger("unoserver")


class UnoServer:
    def __init__(self, interface="127.0.0.1", port="2002"):
        self.interface = interface
        self.port = port

    def start(self, executable="libreoffice"):
        logger.info("Starting unoserver.")

        with tempfile.TemporaryDirectory() as tmpuserdir:

            connection = (
                "socket,host=%s,port=%s,tcpNoDelay=1;urp;StarOffice.ComponentContext"
                % (self.interface, self.port)
            )

            # Store this as an attribute, it helps testing
            self.tmp_uri = "file://" + request.pathname2url(tmpuserdir)

            # I think only --headless and --norestore are needed for
            # command line usage, but let's add everything to be safe.
            cmd = [
                executable,
                "--headless",
                "--invisible",
                "--nocrashreport",
                "--nodefault",
                "--nologo",
                "--nofirststartwizard",
                "--norestore",
                f"-env:UserInstallation={self.tmp_uri}",
                f"--accept={connection}",
            ]

            logger.info("Command: " + " ".join(cmd))
            process = subprocess.Popen(cmd)

            def signal_handler(signum, frame):
                logger.info("Sending signal to LibreOffice")
                try:
                    process.send_signal(signum)
                except ProcessLookupError as e:
                    # 3 means the process is already dead
                    if e.errno != 3:
                        raise

            signal.signal(signal.SIGTERM, signal_handler)
            signal.signal(signal.SIGHUP, signal_handler)
            signal.signal(signal.SIGINT, signal_handler)
            return process


def main():
    logging.basicConfig()
    logger.setLevel(logging.INFO)

    parser = argparse.ArgumentParser("unoserver")
    parser.add_argument(
        "--interface", default="127.0.0.1", help="The interface used by the server"
    )
    parser.add_argument("--port", default="2002", help="The port used by the server")
    parser.add_argument("--daemon", action="store_true", help="Deamonize the server")
    parser.add_argument(
        "--executable",
        default="libreoffice",
        help="The path to the LibreOffice executable",
    )
    args = parser.parse_args()

    server = UnoServer(args.interface, args.port)
    # If it's daemonized, this returns the process.
    # It returns 0 of getting killed in a normal way.
    # Otherwise it returns 1 after the process exits.
    process = server.start(executable=args.executable)
    if args.daemon:
        return process
    pid = process.pid

    process.wait()

    try:
        # Make sure it's really dead
        os.kill(pid, 0)
        # It was killed
        return 0
    except OSError as e:
        if e.errno == 3:
            # All good, it was already dead.
            return 0
        raise


if __name__ == "__main__":
    main()



try:
    import uno
except ImportError:
    raise ImportError(
        "Could not find the 'uno' library. This package must be installed with a Python "
        "installation that has a 'uno' library. This typically means you should install"
        "it with the same Python executable as your Libreoffice installation uses."
    )

import argparse
import io
import logging
import os
import sys
import unohelper

from com.sun.star.beans import PropertyValue
from com.sun.star.io import XOutputStream

logger = logging.getLogger("unoserver")

SFX_FILTER_IMPORT = 1
SFX_FILTER_EXPORT = 2
DOC_TYPES = {
    "com.sun.star.sheet.SpreadsheetDocument",
    "com.sun.star.text.TextDocument",
    "com.sun.star.presentation.PresentationDocument",
    "com.sun.star.drawing.DrawingDocument",
    "com.sun.star.sdb.DocumentDataSource",
    "com.sun.star.formula.FormulaProperties",
    "com.sun.star.script.BasicIDE",
    "com.sun.star.text.WebDocument",  # Supposedly deprecated? But still around.
}


def prop2dict(properties):
    return {p.Name: p.Value for p in properties}


def get_doc_type(doc):
    for t in DOC_TYPES:
        if doc.supportsService(t):
            return t

    # LibreOffice opened it, but it's not one of the known document types.
    # This really should only happen if a future version of LibreOffice starts
    # adding document types, which seems unlikely.
    raise RuntimeError(
        "The input document is of an unknown document type. This is probably a bug.\n"
        "Please create an issue at https://github.com/unoconv/unoserver ."
    )


class OutputStream(unohelper.Base, XOutputStream):
    def __init__(self):
        self.buffer = io.BytesIO()

    def closeOutput(self):
        pass

    def writeBytes(self, seq):
        self.buffer.write(seq.value)


class UnoConverter:
    def __init__(self, interface="127.0.0.1", port="2002"):
        logger.info("Starting unoconverter.")

        self.local_context = uno.getComponentContext()
        self.resolver = self.local_context.ServiceManager.createInstanceWithContext(
            "com.sun.star.bridge.UnoUrlResolver", self.local_context
        )
        self.context = self.resolver.resolve(
            f"uno:socket,host={interface},port={port};urp;StarOffice.ComponentContext"
        )
        self.service = self.context.ServiceManager
        self.desktop = self.service.createInstanceWithContext(
            "com.sun.star.frame.Desktop", self.context
        )
        self.filter_service = self.service.createInstanceWithContext(
            "com.sun.star.document.FilterFactory", self.context
        )
        self.type_service = self.service.createInstanceWithContext(
            "com.sun.star.document.TypeDetection", self.context
        )

    def find_filter(self, import_type, export_type):
        # List export filters. You can only search on module, iflags and eflags,
        # so the import and export types we have to test in a loop
        export_filters = self.filter_service.createSubSetEnumerationByQuery(
            "getSortedFilterList():iflags=2"
        )

        while export_filters.hasMoreElements():
            # Filter DocumentService here
            export_filter = prop2dict(export_filters.nextElement())
            if export_filter["DocumentService"] != import_type:
                continue
            if export_filter["Type"] != export_type:
                continue

            # There is only one possible filter per import and export type,
            # so the first one we find is correct
            return export_filter["Name"]

        # No filter found
        return None

    def convert(self, inpath=None, indata=None, outpath=None, convert_to=None):
        """Converts a file from one type to another

        inpath: A path (on the local hard disk) to a file to be converted.

        indata: A byte string containing the file content to be converted.

        outpath: A path (on the local hard disk) to store the result, or None, in which case
                 the content of the converted file will be returned as a byte string.

        convert_to: The extension of the desired file type, ie "pdf", "xlsx", etc.
        """
        if inpath is None and indata is None:
            raise RuntimeError("Nothing to convert.")

        if inpath is not None and indata is not None:
            raise RuntimeError("You can only pass in inpath or indata, not both.")

        if outpath is None and convert_to is None:
            raise RuntimeError(
                "If you don't specify an output path, you must specify a file-type."
            )

        input_props = (PropertyValue(Name="ReadOnly", Value=True),)

        if inpath:
            # TODO: Verify that inpath exists and is openable, and that outdir exists, because uno's
            # exceptions are completely useless!

            # Load the document
            import_path = uno.systemPathToFileUrl(os.path.abspath(inpath))
            # This returned None if the file was locked, I'm hoping the ReadOnly flag avoids that.
            logger.info(f"Opening {inpath}")

        elif indata:
            # The document content is passed in as a byte string
            input_stream = self.service.createInstanceWithContext(
                "com.sun.star.io.SequenceInputStream", self.context
            )
            input_stream.initialize((uno.ByteSequence(indata),))
            input_props += (PropertyValue(Name="InputStream", Value=input_stream),)
            import_path = "private:stream"

        document = self.desktop.loadComponentFromURL(
            import_path, "_default", 0, input_props
        )

        # Now do the conversion
        try:
            # Figure out document type:
            import_type = get_doc_type(document)

            if outpath:
                export_path = uno.systemPathToFileUrl(os.path.abspath(outpath))
            else:
                export_path = "private:stream"

            # Figure out the output type:
            if convert_to:
                export_type = self.type_service.queryTypeByURL(
                    f"file:///dummy.{convert_to}"
                )
            else:
                export_type = self.type_service.queryTypeByURL(export_path)

            if not export_type:
                if convert_to:
                    extension = convert_to
                else:
                    extension = os.path.splitext(outpath)[-1]
                raise RuntimeError(
                    f"Unknown export file type, unknown extension '{extension}'"
                )

            filtername = self.find_filter(import_type, export_type)
            if filtername is None:
                raise RuntimeError(
                    f"Could not find an export filter from {import_type} to {export_type}"
                )

            logger.info(f"Exporting to {outpath}")
            logger.info(f"Using {filtername} export filter")

            output_props = (
                PropertyValue(Name="FilterName", Value=filtername),
                PropertyValue(Name="Overwrite", Value=True),
            )
            if outpath is None:
                output_stream = OutputStream()
                output_props += (
                    PropertyValue(Name="OutputStream", Value=output_stream),
                )
            document.storeToURL(export_path, output_props)

        finally:
            document.close(True)

        if outpath is None:
            return output_stream.buffer.getvalue()
        else:
            return None


def main():
    logging.basicConfig()
    logger.setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser("unoconvert")
    parser.add_argument(
        "infile", help="The path to the file to be converted (use - for stdin)"
    )
    parser.add_argument(
        "outfile", help="The path to the converted file (use - for stdout)"
    )
    parser.add_argument(
        "--convert-to",
        help="The file type/extension of the output file (ex pdf). Required when using stdout",
    )
    parser.add_argument(
        "--interface", default="127.0.0.1", help="The interface used by the server"
    )
    parser.add_argument("--port", default="2002", help="The port used by the server")
    args = parser.parse_args()

    converter = UnoConverter(args.interface, args.port)

    if args.outfile == "-":
        # Set outfile to None, to get the data returned from the function,
        # instead of written to a file.
        args.outfile = None

    if args.infile == "-":
        # Get data from stdin
        indata = sys.stdin.buffer.read()
        result = converter.convert(
            indata=indata, outpath=args.outfile, convert_to=args.convert_to
        )
    else:
        result = converter.convert(
            inpath=args.infile, outpath=args.outfile, convert_to=args.convert_to
        )

    if args.outfile is None:
        # Pipe result to stdout
        sys.stdout.buffer.write(result)


if __name__ == "__main__":
    main()
